/*	Copyright (C) 1992 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "rtree.h"
#define MAXINT 0x7fffffff

#define POOL 1

struct pool_header
{
  int alloced;
  int pooled;
  void *first;
};

static struct pool_header *pools = 0;
static int biggest_pool = 0;

static struct pool_header *
pool_header_hard (n)
     int n;
{
  if (!biggest_pool)
    {
      biggest_pool = n;
      pools = (struct pool_header *) ck_calloc (sizeof (*pools) * n);
    }

  if (n > biggest_pool)
    {
      pools = (struct pool_header *) ck_realloc (pools, sizeof (*pools) * n);
      bzero (pools + biggest_pool, (n - biggest_pool) * (sizeof (*pools)));
      biggest_pool = n;
    }

  return pools + n - 1;
}

#define pool_header_get(N) ((N) > biggest_pool \
			?pool_header_hard(N) : pools + N - 1)
static void *
ck_palloc (x)
     int x;
{
  struct pool_header *h = pool_header_get (x);
  void *r = h->first;
  ++h->alloced;
  if (r)
    {
      --h->pooled;
      h->first = *(void **) h->first;
    }
  else
    r = ck_malloc (x);

  return r;
}


static void
ck_pfree (m, x)
     void *m;
     int x;
{
  struct pool_header *h = pool_header_get (x);

  h->alloced--;
  if (h->pooled >= 2 * h->alloced)
    {
      while (h->pooled > h->alloced)
	{
	  void *t = h->first;
	  h->first = *(void **) h->first;
	  free (t);
	  --h->pooled;
	}
    }
  ++h->pooled;
  *(void **) m = h->first;
  h->first = m;
}


void *
ck_pcalloc (x)
     int x;
{
  void *r = ck_palloc (x);
  bzero (r, x);
  return r;
}


static xx_RtreeNode
xx_rtreenode ()
{
  return (xx_RtreeNode) ck_pcalloc (sizeof (struct xx_sRtreeNode));
}

xx_RtreeLeaf
xx_rtreeleaf (data)
     void *data;
{
  xx_RtreeLeaf nw = (xx_RtreeLeaf) ck_pcalloc (sizeof (struct xx_sRtreeLeaf));
  nw->data = data;
  return nw;
}

void
xx_free_rtreeleaf (leaf)
     xx_RtreeLeaf leaf;
{
  ck_pfree (leaf, sizeof (*leaf));
}

xx_Rtree
xx_rtree ()
{
  xx_Rtree tree = (xx_Rtree) ck_palloc (sizeof (*tree));
  tree->depth = 1;
  tree->domain.x = tree->domain.y = tree->domain.w = tree->domain.h = 0;
  tree->root = xx_rtreenode ();
  return tree;
}

int
xx_rtree_empty (r)
     xx_Rtree r;
{
  return !r->root->children[0];
}
static void xx_free_rtreenode ();

void
xx_rtree_clear (r)
     xx_Rtree r;
{
  xx_free_rtreenode (r->root, r->depth);
  r->root = xx_rtreenode ();
  r->depth = 1;
}

static void
xx_free_rtreenode (node, depth)
     xx_RtreeNode node;
     int depth;
{
  if (depth - 1)
    {
      int x;
      for (x = 0; xx_rtreenode_subtree (node, x); ++x)
	xx_free_rtreenode (xx_rtreenode_subtree (node, x), depth - 1);
    }
  else
    {
      int x;
      for (x = 0; xx_rtreenode_leaf (node, x); ++x)
	xx_free_rtreeleaf (xx_rtreenode_leaf (node, x));
    }
  ck_pfree (node, sizeof (*node));
}

void
xx_free_rtree (tree)
     xx_Rtree tree;
{
  xx_free_rtreenode (tree->root, tree->depth);
  ck_pfree (tree, sizeof (*tree));
}

static int
xx_rtreenode_query (node, rect, loopfn, loopdata, depth)
     xx_RtreeNode node;
     xx_IntRectangle rect;
     xx_LoopFn loopfn;
     void *loopdata;
     int depth;
{
  if (depth - 1)
    {
      int x;
      for (x = 0; xx_rtreenode_subtree (node, x); ++x)
	{
	  register xx_IntRectangle node_rect = &node->rectangles[x];
	  if (xx_IRintersects (rect, node_rect))
	    if (!xx_rtreenode_query (xx_rtreenode_subtree (node, x), rect,
				     loopfn, loopdata, depth - 1))
	      return 0;
	}
    }
  else
    {
      int x;
      for (x = 0; xx_rtreenode_leaf (node, x); ++x)
	{
	  xx_IntRectangle node_rect = &node->rectangles[x];
	  if (xx_IRintersects (rect, node_rect))
	    if (!loopfn (xx_rtreenode_leaf (node, x), node_rect, loopdata))
	      return 0;
	}
    }

  return 1;
}

void
xx_rtree_query (tree, rect, loopfn, loopdata)
     xx_Rtree tree;
     xx_IntRectangle rect;
     xx_LoopFn loopfn;
     void *loopdata;
{
  xx_rtreenode_query (tree->root, rect, loopfn, loopdata, tree->depth);
}

#define MAX(A,B) ((A) > (B) ? (A) : (B))
#define ABS(A) ((A) < 0 ? -(A) : (A))

static int
xx_normalized_distance (r1, r2)
     xx_IntRectangle r1;
     xx_IntRectangle r2;
{
  int h1 = r1->x - (r2->x + r2->w);
  int h2 = r2->x - (r1->x + r1->w);
  int v1 = r1->y - (r2->y + r2->h);
  int v2 = r2->y - (r1->y + r1->h);

  h1 = ABS (h1);
  h2 = ABS (h2);
  v1 = ABS (v1);
  v2 = ABS (v2);
  v1 = MAX (v1, v2);
  h1 = MAX (h1, h2);
  return MAX (v1, h1);
}


static int
xx_rtreenode_insertchild (newsibling, newsibrect, newrect,
			  node, rect, item, x)
     xx_RtreeNode *newsibling;
     xx_IntRectangle newsibrect;
     xx_IntRectangle newrect;
     xx_RtreeNode node;
     xx_IntRectangle rect;
     xx_RtreeItem item;
     int x;
{
  if (x < RTREE_FANOUT)
    {
      item->parent = node;
      item->pos = x;
      node->children[x] = item;
      node->rectangles[x] = *rect;
      return 0;
    }

  else
    /* x >= RTREE_FANOUT */
    {
      int a, b, i;
      int distance;

      xx_RtreeNode nw = xx_rtreenode ();

      a = 0;
      b = 1;
      distance = xx_normalized_distance (&node->rectangles[a],
					 &node->rectangles[b]);

      for (i = 2; i < RTREE_FANOUT; ++i)
	{
	  int newdist_a = xx_normalized_distance (&node->rectangles[a],
						  &node->rectangles[i]);
	  int newdist_b = xx_normalized_distance (&node->rectangles[b],
						  &node->rectangles[i]);

	  int newdist = MAX (newdist_a, newdist_b);

	  if (newdist > distance)
	    {
	      distance = newdist;
	      if (newdist == newdist_a)
		b = i;
	      else
		a = i;
	    }
	}

      nw->rectangles[0] = node->rectangles[b];
      nw->children[0] = node->children[b];
      xx_rtreenode_item (nw, 0)->parent = nw;
      xx_rtreenode_item (nw, 0)->pos = 0;

      node->children[b] = item;
      node->rectangles[b] = *rect;
      item->parent = node;
      item->pos = b;


      {
	struct xx_sIntRectangle old_area;
	struct xx_sIntRectangle test_old_area;
	int old_area_measure;
	int test_old_area_measure;
	struct xx_sIntRectangle new_area;
	struct xx_sIntRectangle test_new_area;
	int new_area_measure;
	int test_new_area_measure;
	int newpos;

	old_area = node->rectangles[a];
	old_area_measure = xx_IRarea (&old_area);
	new_area = nw->rectangles[0];
	new_area_measure = xx_IRarea (&new_area);
	newpos = 1;

	for (i = (1 + a) % RTREE_FANOUT; i != a; i = (i + 1) % RTREE_FANOUT)
	  {
	    test_old_area = old_area;
	    xx_IRbound (&test_old_area, &node->rectangles[i]);
	    test_old_area_measure = xx_IRarea (&test_old_area);
	    test_new_area = new_area;
	    xx_IRbound (&test_new_area, &node->rectangles[i]);
	    test_new_area_measure = xx_IRarea (&test_new_area);

	    if (test_new_area_measure - new_area_measure
		< test_old_area_measure - old_area_measure)
	      {
		nw->children[newpos] = node->children[i];
		node->children[i] = 0;
		nw->rectangles[newpos] = node->rectangles[i];
		xx_rtreenode_item (nw, newpos)->parent = nw;
		xx_rtreenode_item (nw, newpos)->pos = newpos;
		new_area = test_new_area;
		new_area_measure = test_new_area_measure;
		++newpos;
	      }
	    else
	      {
		old_area = test_old_area;
		old_area_measure = test_old_area_measure;
	      }
	  }

	for (i = 0, newpos = 0; i < RTREE_FANOUT; ++i)
	  {
	    if (node->children[i])
	      {
		node->children[newpos] = node->children[i];
		((xx_RtreeItem) node->children[newpos])->pos = newpos;
		node->rectangles[newpos] = node->rectangles[i];
		++newpos;
	      }
	  }

	while (newpos < RTREE_FANOUT)
	  {
	    node->children[newpos] = 0;
	    ++newpos;
	  }
	*newrect = old_area;
	*newsibrect = new_area;
      }
      *newsibling = nw;
      return 1;
    }
}



static int
xx_rtreenode_add (newnode, newnoderect, newrect, node, rect, item, depth)
     xx_RtreeNode *newnode;
     xx_IntRectangle newnoderect;
     xx_IntRectangle newrect;
     xx_RtreeNode node;
     xx_IntRectangle rect;
     xx_RtreeItem item;
     int depth;
{
  if (!(depth - 1))
    {
      int x;
      for (x = 0; x < RTREE_FANOUT && node->children[x]; ++x)
	;

      return xx_rtreenode_insertchild (newnode, newnoderect, newrect,
				       node, rect, item, x);
    }
  else
    {
      struct xx_sIntRectangle bound, bestbound;
      xx_RtreeNode newchild;
      struct xx_sIntRectangle newchildrect;

      int increase = MAXINT;
      int bestx = 0;
      int x;
      for (x = 0; x < RTREE_FANOUT && node->children[x]; ++x)
	{
	  int area;
	  bound = node->rectangles[x];
	  xx_IRbound (&bound, rect);
	  area = xx_IRarea (&bound);
	  if (area < increase)
	    {
	      bestx = x;
	      bestbound = bound;
	      increase = area;
	    }
	}

      node->rectangles[bestx] = bestbound;
      if (xx_rtreenode_add (&newchild, &newchildrect, &node->rectangles[bestx],
			    xx_rtreenode_subtree (node, bestx),
			    rect, item, depth - 1))
	return xx_rtreenode_insertchild (newnode, newnoderect, newrect,
					 node, &newchildrect,
					 (xx_RtreeItem)newchild, x);
    }
  return 0;
}

void
xx_rtree_add (tree, rect, item)
     xx_Rtree tree;
     xx_IntRectangle rect;
     xx_RtreeLeaf item;
{
  xx_RtreeNode newrootsibling;
  struct xx_sIntRectangle newrootsibling_rect;
  struct xx_sIntRectangle oldroot_newrect;

  if (tree->root->children[0])
    xx_IRbound (&tree->domain, rect);
  else
    tree->domain = *rect;

  if (xx_rtreenode_add (&newrootsibling, &newrootsibling_rect,
			&oldroot_newrect, tree->root,
			rect, (xx_RtreeItem)item, tree->depth))
    {
      xx_RtreeNode nw = xx_rtreenode ();
      nw->children[0] = tree->root;
      nw->rectangles[0] = oldroot_newrect;
      tree->root->parent = nw;
      tree->root->pos = 0;
      nw->children[1] = newrootsibling;
      nw->rectangles[1] = newrootsibling_rect;
      newrootsibling->parent = nw;
      newrootsibling->pos = 1;
      tree->root = nw;
      tree->depth += 1;
    }
}

void
xx_rtree_delete (tree, item)
     xx_Rtree tree;
     xx_RtreeLeaf item;
{
  int reinsertc = 0;
  xx_RtreeItem reinserts[32];
  struct xx_sIntRectangle reinsert_rectangles[32];
  xx_RtreeNode subtree = item->parent;
  int pos = item->pos;

  while (subtree != tree->root && !xx_rtreenode_item (subtree, 2))
    {
      xx_RtreeNode temp = subtree;
      if (subtree->children[1])
	{
	  reinserts[reinsertc] = xx_rtreenode_item (subtree, !pos);
	  reinsert_rectangles[reinsertc] = subtree->rectangles[!pos];
	  ++reinsertc;
	}
      else
	{
	  reinserts[reinsertc] = 0;
	  ++reinsertc;
	}
      pos = subtree->pos;
      subtree = subtree->parent;
      ck_pfree (temp, sizeof (*temp));
      /* xx_free_rtreenode would be wrong here */
    }

  item->parent = 0;
  item->pos = 0;

  if (subtree == tree->root)
    if (!subtree->children[2] && tree->depth > 1)
      {
	xx_RtreeNode livechild = xx_rtreenode_subtree (tree->root, !pos);
	--tree->depth;
	tree->domain = tree->root->rectangles[!pos];
	ck_pfree (tree->root, sizeof (*tree->root));
	tree->root = livechild;
	livechild->parent = 0;
	subtree = 0;
      }

  if (subtree)
    {
      int i;
      for (i = pos + 1; i < RTREE_FANOUT && subtree->children[i]; ++i)
	{
	  subtree->children[i - 1] = subtree->children[i];
	  ((xx_RtreeItem) subtree->children[i - 1])->pos = i - 1;
	  subtree->rectangles[i - 1] = subtree->rectangles[i];
	}
      for (i = i - 1; i < RTREE_FANOUT; ++i)
	subtree->children[i] = 0;

      while (subtree)
	{
	  xx_IntRectangle area = (subtree->parent
				? &subtree->parent->rectangles[subtree->pos]
				  : &tree->domain);

	  *area = subtree->rectangles[0];
	  for (i = 1; i < RTREE_FANOUT && subtree->children[i]; ++i)
	    xx_IRbound (area, &subtree->rectangles[i]);
	  subtree = subtree->parent;
	}
    }

  {
    int i;
    for (i = 0; i < reinsertc; ++i)
      {
	if (reinserts[i])
	  {
	    xx_rtree_add (tree, &reinsert_rectangles[i],
			  (xx_RtreeLeaf) reinserts[i]);
	  }
	--tree->depth;
      }
    tree->depth += i;
  }
  {
    int x;
    if (tree->root->children[0])
      {
	tree->domain = tree->root->rectangles[0];
	for (x = 1; x < RTREE_FANOUT && tree->root->children[x]; ++x)
	  xx_IRbound (&tree->domain, &tree->root->rectangles[x]);
      }
    else
      tree->domain.x = tree->domain.y = tree->domain.h = tree->domain.w = 0;
  }
}

static xx_RtreeNode
xx_rtreenode_copy (node, copier, depth)
     xx_RtreeItem node;
     xx_LoopFn copier;
     int depth;
{
  if (depth - 1)
    {
      xx_RtreeNode n = (xx_RtreeNode) node;
      int x;
      xx_RtreeNode nw = (xx_RtreeNode) ck_pcalloc (sizeof (*nw));
      nw->parent = 0;
      for (x = 0; x < RTREE_FANOUT && n->children[x]; ++x)
	{
	  nw->rectangles[x] = n->rectangles[x];
	  nw->children[x] =
	    xx_rtreenode_copy (n->children[x], copier, depth - 1);
	  ((xx_RtreeItem) nw->children[x])->parent = nw;
	  ((xx_RtreeItem) nw->children[x])->pos = x;
	}
      return nw;
    }
  else
    {
      xx_RtreeNode n = (xx_RtreeNode) node;
      int x;
      xx_RtreeNode nw = (xx_RtreeNode) ck_pcalloc (sizeof (*nw));
      nw->parent = 0;
      for (x = 0; x < RTREE_FANOUT && n->children[x]; ++x)
	{
	  nw->rectangles[x] = n->rectangles[x];
	  (void) copier ((xx_RtreeLeaf) n->children[x],
			 &n->rectangles[x],
			 &nw->children[x]);
	  ((xx_RtreeItem) nw->children[x])->parent = nw;
	  ((xx_RtreeItem) nw->children[x])->pos = x;
	}
      return nw;
    }
}

xx_Rtree
xx_rtree_copy (r, leaf_copier)
     xx_Rtree r;
     xx_LoopFn leaf_copier;
{
  xx_Rtree nw = (xx_Rtree) ck_palloc (sizeof (*nw));
  *nw = *r;
  nw->root = xx_rtreenode_copy ((xx_RtreeItem) r->root, leaf_copier, r->depth);
  return nw;
}

static void
translate_node (node, x, y, depth)
     xx_RtreeNode node;
     int x;
     int y;
{
  int i;
  if (!depth)
    return;
  for (i = 0; i < RTREE_FANOUT && node->children[i]; ++i)
    {
      translate_node (node->children[i], x, y, depth - 1);
      node->rectangles[i].x += x;
      node->rectangles[i].y += y;
    }
}

void
xx_rtree_translate (tree, x, y)
     xx_Rtree tree;
     int x;
     int y;
{
  translate_node (tree->root, x, y, tree->depth);
  tree->domain.x += x;
  tree->domain.y += y;
}
